@@ -74,11 +74,6 @@
 static DBusConnection *connection = NULL;
 static GSList *adapter_drivers = NULL;
 
-struct record_list {
-	sdp_list_t *recs;
-	const gchar *addr;
-};
-
 struct session_req {
 	struct btd_adapter	*adapter;
 	DBusConnection		*conn;		/* Connection reference */
@@ -97,7 +92,7 @@ struct btd_adapter {
 	uint16_t dev_id;
 	int up;
 	char *path;			/* adapter object path */
-	char address[18];		/* adapter Bluetooth Address */
+	bdaddr_t bdaddr;		/* adapter Bluetooth Address */
 	guint discov_timeout_id;	/* discoverable timeout id */
 	uint32_t discov_timeout;	/* discoverable time(msec) */
 	uint8_t scan_mode;		/* scan mode: SCAN_DISABLED, SCAN_PAGE, SCAN_INQUIRY */
@@ -358,7 +353,8 @@ int pending_remote_name_cancel(struct btd_adapter *adapter)
 
 	dev = l->data;
 
-	if (hci_read_remote_name_cancel(dd, &dev->bdaddr, 1000) < 0) {
+	if (hci_read_remote_name_cancel(dd, &dev->bdaddr,
+					HCI_REQ_TIMEOUT) < 0) {
 		error("Remote name cancel failed: %s(%d)", strerror(errno), errno);
 		err = -errno;
 	}
@@ -463,7 +459,7 @@ static const char *mode2str(uint8_t mode)
 	}
 }
 
-static uint8_t str2mode(const char *addr, const char *mode)
+static uint8_t get_mode(const bdaddr_t *bdaddr, const char *mode)
 {
 	if (strcasecmp("off", mode) == 0)
 		return MODE_OFF;
@@ -474,11 +470,13 @@ static uint8_t str2mode(const char *addr, const char *mode)
 	else if (strcasecmp("limited", mode) == 0)
 		return MODE_LIMITED;
 	else if (strcasecmp("on", mode) == 0) {
-		char onmode[14];
-		if (read_on_mode(addr, onmode, sizeof(onmode)) < 0)
+		char onmode[14], srcaddr[18];
+
+		ba2str(bdaddr, srcaddr);
+		if (read_on_mode(srcaddr, onmode, sizeof(onmode)) < 0)
 			return MODE_CONNECTABLE;
 
-		return str2mode(addr, onmode);
+		return get_mode(bdaddr, onmode);
 	} else
 		return MODE_UNKNOWN;
 }
@@ -489,7 +487,6 @@ static DBusMessage *set_mode(DBusConnection *conn, DBusMessage *msg,
 	struct btd_adapter *adapter = data;
 	uint8_t scan_enable;
 	uint8_t current_scan = adapter->scan_mode;
-	bdaddr_t local;
 	gboolean limited;
 	int err, dd;
 	const char *mode;
@@ -564,7 +561,7 @@ static DBusMessage *set_mode(DBusConnection *conn, DBusMessage *msg,
 		rq.rlen   = sizeof(status);
 		rq.event = EVT_CMD_COMPLETE;
 
-		if (hci_send_req(dd, &rq, 1000) < 0) {
+		if (hci_send_req(dd, &rq, HCI_REQ_TIMEOUT) < 0) {
 			err = errno;
 			error("Sending write scan enable command failed: %s (%d)",
 					strerror(errno), errno);
@@ -592,8 +589,7 @@ static DBusMessage *set_mode(DBusConnection *conn, DBusMessage *msg,
 		}
 	}
 done:
-	str2ba(adapter->address, &local);
-	write_device_mode(&local, mode);
+	write_device_mode(&adapter->bdaddr, mode);
 
 	hci_close_dev(dd);
 
@@ -602,6 +598,36 @@ done:
 	return dbus_message_new_method_return(msg);
 }
 
+static DBusMessage *set_powered(DBusConnection *conn, DBusMessage *msg,
+				gboolean powered, void *data)
+{
+	struct btd_adapter *adapter = data;
+	uint8_t mode;
+
+	mode = powered ? get_mode(&adapter->bdaddr, "on") : MODE_OFF;
+
+	if (mode == adapter->mode)
+		return dbus_message_new_method_return(msg);
+
+	return set_mode(conn, msg, mode, data);
+}
+
+static DBusMessage *set_discoverable(DBusConnection *conn, DBusMessage *msg,
+				gboolean discoverable, void *data)
+{
+	struct btd_adapter *adapter = data;
+	const char *strmode;
+	uint8_t mode;
+
+	strmode = discoverable ? "discoverable" : "connectable";
+	mode = get_mode(&adapter->bdaddr, strmode);
+
+	if (mode == adapter->mode)
+		return dbus_message_new_method_return(msg);
+
+	return set_mode(conn, msg, mode, data);
+}
+
 static struct session_req *find_session(GSList *list, DBusMessage *msg)
 {
 	GSList *l;
@@ -758,7 +784,7 @@ static DBusMessage *confirm_mode(DBusConnection *conn, DBusMessage *msg,
 	if (!adapter->agent)
 		return dbus_message_new_method_return(msg);
 
-	umode = str2mode(adapter->address, mode);
+	umode = get_mode(&adapter->bdaddr, mode);
 
 	req = create_session(adapter, conn, msg, umode, NULL);
 
@@ -778,7 +804,6 @@ static DBusMessage *set_discoverable_timeout(DBusConnection *conn,
 							void *data)
 {
 	struct btd_adapter *adapter = data;
-	bdaddr_t bdaddr;
 	const char *path;
 
 	if (adapter->discov_timeout_id) {
@@ -791,8 +816,7 @@ static DBusMessage *set_discoverable_timeout(DBusConnection *conn,
 
 	adapter->discov_timeout = timeout;
 
-	str2ba(adapter->address, &bdaddr);
-	write_discoverable_timeout(&bdaddr, timeout);
+	write_discoverable_timeout(&adapter->bdaddr, timeout);
 
 	path = dbus_message_get_path(msg);
 
@@ -816,7 +840,8 @@ static void update_ext_inquiry_response(int dd, struct hci_dev *dev)
 	if (dev->ssp_mode > 0)
 		create_ext_inquiry_response((char *) dev->name, data);
 
-	if (hci_write_ext_inquiry_response(dd, fec, data, 2000) < 0)
+	if (hci_write_ext_inquiry_response(dd, fec, data,
+						HCI_REQ_TIMEOUT) < 0)
 		error("Can't write extended inquiry response: %s (%d)",
 						strerror(errno), errno);
 }
@@ -825,11 +850,8 @@ static int adapter_set_name(struct btd_adapter *adapter, const char *name)
 {
 	struct hci_dev *dev = &adapter->dev;
 	int dd, err;
-	bdaddr_t bdaddr;
 
-	str2ba(adapter->address, &bdaddr);
-
-	write_local_name(&bdaddr, (char *) name);
+	write_local_name(&adapter->bdaddr, (char *) name);
 
 	if (!adapter->up)
 		return 0;
@@ -842,7 +864,7 @@ static int adapter_set_name(struct btd_adapter *adapter, const char *name)
 		return -err;
 	}
 
-	if (hci_write_local_name(dd, name, 5000) < 0) {
+	if (hci_write_local_name(dd, name, HCI_REQ_TIMEOUT) < 0) {
 		err = errno;
 		error("Can't write name for hci%d: %s (%d)",
 					adapter->dev_id, strerror(err), err);
@@ -942,20 +964,20 @@ static DBusMessage *remove_bonding(DBusConnection *conn, DBusMessage *msg,
 	struct btd_adapter *adapter = data;
 	struct btd_device *device;
 	char filename[PATH_MAX + 1];
-	char *str;
-	bdaddr_t src, dst;
+	char *str, srcaddr[18];
+	bdaddr_t dst;
 	GSList *l;
 	int dev, err;
 	gboolean paired;
 
-	str2ba(adapter->address, &src);
 	str2ba(address, &dst);
+	ba2str(&adapter->bdaddr, srcaddr);
 
 	dev = hci_open_dev(adapter->dev_id);
 	if (dev < 0 && msg)
 		return no_such_adapter(msg);
 
-	create_name(filename, PATH_MAX, STORAGEDIR, adapter->address,
+	create_name(filename, PATH_MAX, STORAGEDIR, srcaddr,
 			"linkkeys");
 
 	/* textfile_del doesn't return an error when the key is not found */
@@ -978,7 +1000,7 @@ static DBusMessage *remove_bonding(DBusConnection *conn, DBusMessage *msg,
 	}
 
 	/* Delete the link key from the Bluetooth chip */
-	hci_delete_stored_link_key(dev, &dst, 0, 1000);
+	hci_delete_stored_link_key(dev, &dst, 0, HCI_REQ_TIMEOUT);
 
 	/* find the connection */
 	l = g_slist_find_custom(adapter->active_conn, &dst,
@@ -987,7 +1009,8 @@ static DBusMessage *remove_bonding(DBusConnection *conn, DBusMessage *msg,
 		struct active_conn_info *con = l->data;
 		/* Send the HCI disconnect command */
 		if ((hci_disconnect(dev, htobs(con->handle),
-					HCI_OE_USER_ENDED_CONNECTION, 500) < 0)
+					HCI_OE_USER_ENDED_CONNECTION,
+					HCI_REQ_TIMEOUT) < 0)
 					&& msg){
 			int err = errno;
 			error("Disconnect failed");
@@ -1026,25 +1049,27 @@ done:
 void adapter_remove_device(DBusConnection *conn, struct btd_adapter *adapter,
 				struct btd_device *device)
 {
-	bdaddr_t src, dst;
+	bdaddr_t dst;
 	const gchar *dev_path = device_get_path(device);
 	struct agent *agent;
-	char dst_addr[18];
+	char dstaddr[18];
 
 	device_get_address(device, &dst);
-	ba2str(&dst, dst_addr);
+	ba2str(&dst, dstaddr);
 
-	str2ba(adapter->address, &src);
-	delete_entry(&src, "profiles", dst_addr);
+	delete_entry(&adapter->bdaddr, "profiles", dstaddr);
+	adapter->devices = g_slist_remove(adapter->devices, device);
 
 	if (!device_is_temporary(device)) {
-		remove_bonding(conn, NULL, dst_addr, adapter);
+		remove_bonding(conn, NULL, dstaddr, adapter);
 
 		g_dbus_emit_signal(conn, adapter->path,
 				ADAPTER_INTERFACE,
 				"DeviceRemoved",
 				DBUS_TYPE_OBJECT_PATH, &dev_path,
 				DBUS_TYPE_INVALID);
+
+		adapter_update_devices(adapter);
 	}
 
 	agent = device_get_agent(device);
@@ -1054,8 +1079,6 @@ void adapter_remove_device(DBusConnection *conn, struct btd_adapter *adapter,
 		device_set_agent(device, NULL);
 	}
 
-	adapter->devices = g_slist_remove(adapter->devices, device);
-
 	device_remove(conn, device);
 }
 
@@ -1104,7 +1127,6 @@ static gboolean create_bonding_conn_complete(GIOChannel *io, GIOCondition cond,
 		/* If we come here it implies a bug somewhere */
 		debug("create_bonding_conn_complete: no pending bonding!");
 		g_io_channel_close(io);
-		g_io_channel_unref(io);
 		return FALSE;
 	}
 
@@ -1179,7 +1201,7 @@ static gboolean create_bonding_conn_complete(GIOChannel *io, GIOCondition cond,
 	rq.rlen   = EVT_CMD_STATUS_SIZE;
 	rq.event  = EVT_CMD_STATUS;
 
-	if (hci_send_req(dd, &rq, 500) < 0) {
+	if (hci_send_req(dd, &rq, HCI_REQ_TIMEOUT) < 0) {
 		error("Unable to send HCI request: %s (%d)",
 					strerror(errno), errno);
 		error_failed_errno(adapter->bonding->conn, adapter->bonding->msg,
@@ -1257,28 +1279,43 @@ static void cancel_auth_request(struct pending_auth_info *auth, int dev_id)
 	hci_close_dev(dd);
 }
 
-static void create_bond_req_exit(void *user_data)
+static void cancel_bonding(struct btd_adapter *adapter, gboolean exited)
 {
-	struct btd_adapter *adapter = user_data;
 	struct pending_auth_info *auth;
+	struct bonding_request_info *bonding = adapter->bonding;
 
-	debug("CreateConnection requestor exited before bonding was completed");
-
-	auth = adapter_find_auth_request(adapter, &adapter->bonding->bdaddr);
+	auth = adapter_find_auth_request(adapter, &adapter->bdaddr);
 	if (auth) {
 		cancel_auth_request(auth, adapter->dev_id);
 		if (auth->agent)
 			agent_cancel(auth->agent);
-		adapter_remove_auth_request(adapter, &adapter->bonding->bdaddr);
+		adapter_remove_auth_request(adapter, &bonding->bdaddr);
 	}
 
 	remove_pending_device(adapter);
 
-	g_io_channel_close(adapter->bonding->io);
-	if (adapter->bonding->io_id)
-		g_source_remove(adapter->bonding->io_id);
-	bonding_request_free(adapter->bonding);
-	adapter->bonding = NULL;
+	if (bonding->io)
+		g_io_channel_close(bonding->io);
+
+	if (exited) {
+		if (bonding->io_id) {
+			g_source_remove(bonding->io_id);
+			bonding->io_id = 0;
+		}
+		bonding_request_free(bonding);
+		adapter->bonding = NULL;
+
+	} else
+		bonding->cancel = TRUE;
+}
+
+static void create_bond_req_exit(void *user_data)
+{
+	struct btd_adapter *adapter = user_data;
+
+	debug("CreateConnection requestor exited before bonding was completed");
+
+	cancel_bonding(adapter, TRUE);
 }
 
 static DBusMessage *create_bonding(DBusConnection *conn, DBusMessage *msg,
@@ -1286,13 +1323,14 @@ static DBusMessage *create_bonding(DBusConnection *conn, DBusMessage *msg,
 				uint8_t capability, void *data)
 {
 	char filename[PATH_MAX + 1];
-	char *str;
+	char *str, srcaddr[18];
 	struct btd_adapter *adapter = data;
 	struct bonding_request_info *bonding;
-	bdaddr_t bdaddr;
+	bdaddr_t dst;
 	int sk;
 
-	str2ba(address, &bdaddr);
+	str2ba(address, &dst);
+	ba2str(&adapter->bdaddr, srcaddr);
 
 	/* check if there is a pending discover: requested by D-Bus/non clients */
 	if (adapter->state & STD_INQUIRY)
@@ -1303,11 +1341,11 @@ static DBusMessage *create_bonding(DBusConnection *conn, DBusMessage *msg,
 	if (adapter->bonding)
 		return in_progress(msg, "Bonding in progress");
 
-	if (adapter_find_auth_request(adapter, &bdaddr))
+	if (adapter_find_auth_request(adapter, &dst))
 		return in_progress(msg, "Bonding in progress");
 
 	/* check if a link key already exists */
-	create_name(filename, PATH_MAX, STORAGEDIR, adapter->address,
+	create_name(filename, PATH_MAX, STORAGEDIR, srcaddr,
 			"linkkeys");
 
 	str = textfile_caseget(filename, address);
@@ -1318,7 +1356,7 @@ static DBusMessage *create_bonding(DBusConnection *conn, DBusMessage *msg,
 				"Bonding already exists");
 	}
 
-	sk = l2raw_connect(adapter->address, &bdaddr);
+	sk = l2raw_connect(&adapter->bdaddr, &dst);
 	if (sk < 0)
 		return g_dbus_create_error(msg,
 				ERROR_INTERFACE ".ConnectionAttemptFailed",
@@ -1375,7 +1413,7 @@ int start_inquiry(struct btd_adapter *adapter)
 	rq.rlen = EVT_CMD_STATUS_SIZE;
 	rq.event = EVT_CMD_STATUS;
 
-	if (hci_send_req(dd, &rq, 500) < 0) {
+	if (hci_send_req(dd, &rq, HCI_REQ_TIMEOUT) < 0) {
 		err = errno;
 		error("Unable to start inquiry: %s (%d)",
 			strerror(err), err);
@@ -1426,7 +1464,7 @@ static int start_periodic_inquiry(struct btd_adapter *adapter)
 	rq.rlen   = sizeof(status);
 	rq.event  = EVT_CMD_COMPLETE;
 
-	if (hci_send_req(dd, &rq, 1000) < 0) {
+	if (hci_send_req(dd, &rq, HCI_REQ_TIMEOUT) < 0) {
 		err = errno;
 		error("Unable to start periodic inquiry: %s (%d)",
 				strerror(err), err);
@@ -1517,11 +1555,15 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	DBusMessage *reply;
 	DBusMessageIter iter;
 	DBusMessageIter dict;
-	bdaddr_t ba;
-	char str[249];
-	gboolean discov_active;
+	char str[249], srcaddr[18];
+	gboolean bproperty;
+	char **devices;
+	int i;
+	GSList *l;
 
-	if (check_address(adapter->address) < 0)
+	ba2str(&adapter->bdaddr, srcaddr);
+
+	if (check_address(srcaddr) < 0)
 		return adapter_not_ready(msg);
 
 	reply = dbus_message_new_method_return(msg);
@@ -1536,37 +1578,56 @@ static DBusMessage *get_properties(DBusConnection *conn,
 			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
 
 	/* Address */
-	property = adapter->address;
+	property = srcaddr;
 	dbus_message_iter_append_dict_entry(&dict, "Address",
-			DBUS_TYPE_STRING, &property);
+						DBUS_TYPE_STRING, &property);
 
 	/* Name */
 	memset(str, 0, sizeof(str));
+	strncpy(str, (char *) adapter->dev.name, 248);
 	property = str;
-	str2ba(adapter->address, &ba);
 
-	if (!read_local_name(&ba, str))
-		dbus_message_iter_append_dict_entry(&dict, "Name",
-			DBUS_TYPE_STRING, &property);
+	dbus_message_iter_append_dict_entry(&dict, "Name",
+						DBUS_TYPE_STRING, &property);
 
 	/* Mode */
 	property = mode2str(adapter->mode);
 
 	dbus_message_iter_append_dict_entry(&dict, "Mode",
-			DBUS_TYPE_STRING, &property);
+						DBUS_TYPE_STRING, &property);
+
+	/* Powered */
+	bproperty = adapter->up ? TRUE : FALSE;
+	dbus_message_iter_append_dict_entry(&dict, "Powered",
+					DBUS_TYPE_BOOLEAN, &bproperty);
+
+	/* Discoverable */
+	bproperty = adapter->scan_mode == SCAN_INQUIRY ? TRUE : FALSE;
+	dbus_message_iter_append_dict_entry(&dict, "Discoverable",
+					DBUS_TYPE_BOOLEAN, &bproperty);
 
 	/* DiscoverableTimeout */
 	dbus_message_iter_append_dict_entry(&dict, "DiscoverableTimeout",
 				DBUS_TYPE_UINT32, &adapter->discov_timeout);
 
 	if (adapter->state & PERIODIC_INQUIRY || adapter->state & STD_INQUIRY)
-		discov_active = TRUE;
+		bproperty = TRUE;
 	else
-		discov_active = FALSE;
+		bproperty = FALSE;
 
-	/* PeriodicDiscovery */
+	/* Discovering */
 	dbus_message_iter_append_dict_entry(&dict, "Discovering",
-				DBUS_TYPE_BOOLEAN, &discov_active);
+					DBUS_TYPE_BOOLEAN, &bproperty);
+
+	/* Devices */
+	devices = g_new0(char *, g_slist_length(adapter->devices) + 1);
+	for (i = 0, l = adapter->devices; l; l = l->next, i++) {
+		struct btd_device *dev = l->data;
+		devices[i] = (char *) device_get_path(dev);
+	}
+	dbus_message_iter_append_dict_entry(&dict, "Devices",
+			DBUS_TYPE_ARRAY, &devices);
+	g_free(devices);
 
 	dbus_message_iter_close_container(&iter, &dict);
 
@@ -1580,6 +1641,9 @@ static DBusMessage *set_property(DBusConnection *conn,
 	DBusMessageIter iter;
 	DBusMessageIter sub;
 	const char *property;
+	char srcaddr[18];
+
+	ba2str(&adapter->bdaddr, srcaddr);
 
 	if (!dbus_message_iter_init(msg, &iter))
 		return invalid_args(msg);
@@ -1618,7 +1682,7 @@ static DBusMessage *set_property(DBusConnection *conn,
 
 		dbus_message_iter_get_basic(&sub, &mode);
 
-		adapter->global_mode = str2mode(adapter->address, mode);
+		adapter->global_mode = get_mode(&adapter->bdaddr, mode);
 
 		if (adapter->global_mode == adapter->mode)
 			return dbus_message_new_method_return(msg);
@@ -1626,27 +1690,44 @@ static DBusMessage *set_property(DBusConnection *conn,
 		if (adapter->mode_sessions && adapter->global_mode < adapter->mode)
 			return confirm_mode(conn, msg, mode, data);
 
-		return set_mode(conn, msg, str2mode(adapter->address, mode),
+		return set_mode(conn, msg, get_mode(&adapter->bdaddr, mode),
 				data);
+	} else if (g_str_equal("Powered", property)) {
+		gboolean powered;
+
+		if (dbus_message_iter_get_arg_type(&sub) != DBUS_TYPE_BOOLEAN)
+			return invalid_args(msg);
+
+		dbus_message_iter_get_basic(&sub, &powered);
+
+		return set_powered(conn, msg, powered, data);
+	} else if (g_str_equal("Discoverable", property)) {
+		gboolean discoverable;
+
+		if (dbus_message_iter_get_arg_type(&sub) != DBUS_TYPE_BOOLEAN)
+			return invalid_args(msg);
+
+		dbus_message_iter_get_basic(&sub, &discoverable);
+
+		return set_discoverable(conn, msg, discoverable, data);
 	}
 
 	return invalid_args(msg);
 }
 
-static DBusMessage *request_mode(DBusConnection *conn,
-					DBusMessage *msg, void *data)
+static DBusMessage *mode_request(DBusConnection *conn,
+					DBusMessage *msg, const char *mode,
+					void *data)
 {
-	const char *mode;
 	struct btd_adapter *adapter = data;
 	struct session_req *req;
 	uint8_t new_mode;
 	int ret;
+	char srcaddr[18];
 
-	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &mode,
-						DBUS_TYPE_INVALID))
-		return invalid_args(msg);
+	ba2str(&adapter->bdaddr, srcaddr);
 
-	new_mode = str2mode(adapter->address, mode);
+	new_mode = get_mode(&adapter->bdaddr, mode);
 	if (new_mode != MODE_CONNECTABLE && new_mode != MODE_DISCOVERABLE)
 		return invalid_args(msg);
 
@@ -1685,7 +1766,25 @@ static DBusMessage *request_mode(DBusConnection *conn,
 	return NULL;
 }
 
-static DBusMessage *release_mode(DBusConnection *conn,
+static DBusMessage *request_mode(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	const char *mode;
+
+	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &mode,
+						DBUS_TYPE_INVALID))
+		return invalid_args(msg);
+
+	return mode_request(conn, msg, mode, data);
+}
+
+static DBusMessage *request_session(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	return mode_request(conn, msg, "connectable", data);
+}
+
+static DBusMessage *release_session(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
 	struct btd_adapter *adapter = data;
@@ -1739,6 +1838,35 @@ static DBusMessage *list_devices(DBusConnection *conn,
 	return reply;
 }
 
+static DBusMessage *cancel_device_creation(DBusConnection *conn,
+						DBusMessage *msg, void *data)
+{
+	struct btd_adapter *adapter = data;
+	struct bonding_request_info *bonding = adapter->bonding;
+	const gchar *address;
+	bdaddr_t bda;
+
+	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &address,
+						DBUS_TYPE_INVALID) == FALSE)
+		return invalid_args(msg);
+
+	if (check_address(address) < 0)
+		return invalid_args(msg);
+
+	str2ba(address, &bda);
+
+	if (bonding && !bacmp(&bonding->bdaddr, &bda)) {
+		if (!g_str_equal(dbus_message_get_sender(msg),
+					dbus_message_get_sender(bonding->msg)))
+			return not_authorized(msg);
+
+		debug("Canceling device creation for %s", address);
+		cancel_bonding(adapter, FALSE);
+	}
+
+	return dbus_message_new_method_return(msg);
+}
+
 static DBusMessage *create_device(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -1971,16 +2099,23 @@ static GDBusMethodTable adapter_methods[] = {
 	{ "SetProperty",	"sv",	"",	set_property,
 						G_DBUS_METHOD_FLAG_ASYNC},
 	{ "RequestMode",	"s",	"",	request_mode,
+						G_DBUS_METHOD_FLAG_ASYNC |
+						G_DBUS_METHOD_FLAG_DEPRECATED},
+	{ "ReleaseMode",	"",	"",	release_session,
+						G_DBUS_METHOD_FLAG_DEPRECATED},
+	{ "RequestSession",	"",	"",	request_session,
 						G_DBUS_METHOD_FLAG_ASYNC},
-	{ "ReleaseMode",	"",	"",	release_mode		},
+	{ "ReleaseSession",	"",	"",	release_session		},
 	{ "StartDiscovery",	"",	"",	adapter_start_discovery },
 	{ "StopDiscovery",	"",	"",	adapter_stop_discovery,
 						G_DBUS_METHOD_FLAG_ASYNC},
-	{ "ListDevices",	"",	"ao",	list_devices		},
+	{ "ListDevices",	"",	"ao",	list_devices,
+						G_DBUS_METHOD_FLAG_DEPRECATED},
 	{ "CreateDevice",	"s",	"o",	create_device,
 						G_DBUS_METHOD_FLAG_ASYNC},
 	{ "CreatePairedDevice",	"sos",	"o",	create_paired_device,
 						G_DBUS_METHOD_FLAG_ASYNC},
+	{ "CancelDeviceCreation","s",	"",	cancel_device_creation	},
 	{ "RemoveDevice",	"o",	"",	remove_device		},
 	{ "FindDevice",		"s",	"o",	find_device		},
 	{ "RegisterAgent",	"os",	"",	register_agent		},
@@ -2025,7 +2160,7 @@ static inline uint8_t get_inquiry_mode(struct hci_dev *dev)
 	return 0;
 }
 
-static int device_read_bdaddr(uint16_t dev_id, bdaddr_t *bdaddr)
+static int adapter_read_bdaddr(uint16_t dev_id, bdaddr_t *bdaddr)
 {
 	int dd, err;
 
@@ -2037,7 +2172,7 @@ static int device_read_bdaddr(uint16_t dev_id, bdaddr_t *bdaddr)
 		return -err;
 	}
 
-	if (hci_read_bd_addr(dd, bdaddr, 2000) < 0) {
+	if (hci_read_bd_addr(dd, bdaddr, HCI_REQ_TIMEOUT) < 0) {
 		err = errno;
 		error("Can't read address for hci%d: %s (%d)",
 					dev_id, strerror(err), err);
@@ -2055,7 +2190,6 @@ static int adapter_setup(struct btd_adapter *adapter, int dd)
 	struct hci_dev *dev = &adapter->dev;
 	uint8_t events[8] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0x1f, 0x00, 0x00 };
 	uint8_t inqmode;
-	bdaddr_t bdaddr;
 	int err;
 	char name[249];
 
@@ -2091,10 +2225,9 @@ static int adapter_setup(struct btd_adapter *adapter, int dd)
 						sizeof(events), events);
 	}
 
-	str2ba(adapter->address, &bdaddr);
-	if (read_local_name(&bdaddr, name) == 0) {
+	if (read_local_name(&adapter->bdaddr, name) == 0) {
 		memcpy(dev->name, name, 248);
-		hci_write_local_name(dd, name, 5000);
+		hci_write_local_name(dd, name, HCI_REQ_TIMEOUT);
         }
 
 	update_ext_inquiry_response(dd, dev);
@@ -2103,7 +2236,7 @@ static int adapter_setup(struct btd_adapter *adapter, int dd)
 	if (inqmode < 1)
 		return 0;
 
-	if (hci_write_inquiry_mode(dd, inqmode, 2000) < 0) {
+	if (hci_write_inquiry_mode(dd, inqmode, HCI_REQ_TIMEOUT) < 0) {
 		err = errno;
 		error("Can't write inquiry mode for %s: %s (%d)",
 					adapter->path, strerror(err), err);
@@ -2128,44 +2261,17 @@ static int active_conn_append(GSList **list, bdaddr_t *bdaddr,
 	return 0;
 }
 
-static void create_stored_records_from_keys(char *key, char *value,
-						void *user_data)
-{
-	struct record_list *rec_list = user_data;
-	const gchar *addr = rec_list->addr;
-	sdp_record_t *rec;
-	int size, i, len;
-	uint8_t *pdata;
-	char tmp[3] = "";
-
-	if (strstr(key, addr) == NULL)
-		return;
-
-	size = strlen(value)/2;
-	pdata = g_malloc0(size);
-
-	for (i = 0; i < size; i++) {
-		 memcpy(tmp, value + (i * 2), 2);
-		 pdata[i] = (uint8_t) strtol(tmp, NULL, 16);
-	}
-
-	rec = sdp_extract_pdu(pdata, size, &len);
-	free(pdata);
-
-	rec_list->recs = sdp_list_append(rec_list->recs, rec);
-}
-
 static void create_stored_device_from_profiles(char *key, char *value,
 						void *user_data)
 {
-	char filename[PATH_MAX + 1];
 	struct btd_adapter *adapter = user_data;
 	GSList *uuids = bt_string2list(value);
 	struct btd_device *device;
-	const gchar *src;
-	struct record_list rec_list;
+	sdp_list_t *records;
 	bdaddr_t dst;
-	char dst_addr[18];
+	char srcaddr[18], dstaddr[18];
+
+	ba2str(&adapter->bdaddr, srcaddr);
 
 	if (g_slist_find_custom(adapter->devices,
 				key, (GCompareFunc) device_address_cmp))
@@ -2179,19 +2285,14 @@ static void create_stored_device_from_profiles(char *key, char *value,
 	adapter->devices = g_slist_append(adapter->devices, device);
 
 	device_get_address(device, &dst);
-	ba2str(&dst, dst_addr);
+	ba2str(&dst, dstaddr);
 
-	src = adapter->address;
-	rec_list.addr = dst_addr;
-	rec_list.recs = NULL;
+	records = read_records(srcaddr, dstaddr);
 
-	create_name(filename, PATH_MAX, STORAGEDIR, src, "sdp");
-	textfile_foreach(filename, create_stored_records_from_keys, &rec_list);
+	device_probe_drivers(device, uuids, records);
 
-	device_probe_drivers(device, uuids, rec_list.recs);
-
-	if (rec_list.recs != NULL)
-		sdp_list_free(rec_list.recs, (sdp_free_func_t) sdp_record_free);
+	if (records)
+		sdp_list_free(records, (sdp_free_func_t) sdp_record_free);
 
 	g_slist_free(uuids);
 }
@@ -2216,11 +2317,14 @@ static void create_stored_device_from_linkkeys(char *key, char *value,
 static void load_devices(struct btd_adapter *adapter)
 {
 	char filename[PATH_MAX + 1];
+	char srcaddr[18];
 
-	create_name(filename, PATH_MAX, STORAGEDIR, adapter->address, "profiles");
+	ba2str(&adapter->bdaddr, srcaddr);
+
+	create_name(filename, PATH_MAX, STORAGEDIR, srcaddr, "profiles");
 	textfile_foreach(filename, create_stored_device_from_profiles, adapter);
 
-	create_name(filename, PATH_MAX, STORAGEDIR, adapter->address, "linkkeys");
+	create_name(filename, PATH_MAX, STORAGEDIR, srcaddr, "linkkeys");
 	textfile_foreach(filename, create_stored_device_from_linkkeys, adapter);
 }
 
@@ -2251,18 +2355,39 @@ static void adapter_up(struct btd_adapter *adapter, int dd)
 	struct hci_conn_list_req *cl = NULL;
 	struct hci_conn_info *ci;
 	const char *pmode;
-	char mode[14];
+	char mode[14], srcaddr[18];
 	int i;
+	gboolean powered;
+	gboolean discoverable;
+
+	ba2str(&adapter->bdaddr, srcaddr);
 
 	adapter->up = 1;
-	adapter->discov_timeout = get_discoverable_timeout(adapter->address);
+	adapter->discov_timeout = get_discoverable_timeout(srcaddr);
 	adapter->state = DISCOVER_TYPE_NONE;
 
 	/* Set scan mode */
-	if (read_device_mode(adapter->address, mode, sizeof(mode)) == 0) {
-		if (!strcmp(mode, "off") && main_opts.offmode == HCID_OFFMODE_NOSCAN) {
-			adapter->mode = MODE_OFF;
-			adapter->scan_mode= SCAN_DISABLED;
+	if (read_device_mode(srcaddr, mode, sizeof(mode)) == 0) {
+		if (!strcmp(mode, "off")) {
+			if (main_opts.offmode == HCID_OFFMODE_NOSCAN) {
+				adapter->mode = MODE_OFF;
+				adapter->scan_mode= SCAN_DISABLED;
+			} else if (main_opts.offmode == HCID_OFFMODE_DEVDOWN) {
+				static gboolean restore_on_mode = FALSE;
+
+				if (!restore_on_mode) {
+					ioctl(dd, HCIDEVDOWN, adapter->dev_id);
+					restore_on_mode = TRUE;
+					return;
+				}
+
+				if (read_on_mode(srcaddr, mode, sizeof(mode)) < 0)
+					write_device_mode(&adapter->bdaddr, mode);
+				else
+					write_device_mode(&adapter->bdaddr, "connectable");
+
+				adapter_up(adapter, dd);
+			}
 		} else if (!strcmp(mode, "connectable")) {
 			adapter->mode = MODE_CONNECTABLE;
 			adapter->scan_mode = SCAN_PAGE;
@@ -2318,6 +2443,19 @@ static void adapter_up(struct btd_adapter *adapter, int dd)
 					ADAPTER_INTERFACE, "Mode",
 					DBUS_TYPE_STRING, &pmode);
 
+	powered = adapter->scan_mode == SCAN_DISABLED ? FALSE : TRUE;
+
+	dbus_connection_emit_property_changed(connection, adapter->path,
+					ADAPTER_INTERFACE, "Powered",
+					DBUS_TYPE_BOOLEAN, &powered);
+
+	discoverable = adapter->scan_mode == (SCAN_PAGE | SCAN_INQUIRY) ? TRUE
+				: FALSE;
+
+	dbus_connection_emit_property_changed(connection, adapter->path,
+					ADAPTER_INTERFACE, "Discoverable",
+					DBUS_TYPE_BOOLEAN, &discoverable);
+
 	load_drivers(adapter);
 	load_devices(adapter);
 }
@@ -2344,12 +2482,12 @@ int adapter_start(struct btd_adapter *adapter)
 
 		debug("Adapter %s without an address", adapter->path);
 
-		err = device_read_bdaddr(adapter->dev_id, &di.bdaddr);
+		err = adapter_read_bdaddr(adapter->dev_id, &di.bdaddr);
 		if (err < 0)
 			return err;
 	}
 
-	ba2str(&di.bdaddr, adapter->address);
+	bacpy(&adapter->bdaddr, &di.bdaddr);
 	memcpy(dev->features, di.features, 8);
 
 	dd = hci_open_dev(adapter->dev_id);
@@ -2360,7 +2498,7 @@ int adapter_start(struct btd_adapter *adapter)
 		return -err;
 	}
 
-	if (hci_read_local_version(dd, &ver, 1000) < 0) {
+	if (hci_read_local_version(dd, &ver, HCI_REQ_TIMEOUT) < 0) {
 		err = errno;
 		error("Can't read version info for %s: %s (%d)",
 					adapter->path, strerror(err), err);
@@ -2373,7 +2511,7 @@ int adapter_start(struct btd_adapter *adapter)
 	dev->lmp_subver = ver.lmp_subver;
 	dev->manufacturer = ver.manufacturer;
 
-	if (hci_read_local_features(dd, features, 1000) < 0) {
+	if (hci_read_local_features(dd, features, HCI_REQ_TIMEOUT) < 0) {
 		err = errno;
 		error("Can't read features for %s: %s (%d)",
 					adapter->path, strerror(err), err);
@@ -2383,7 +2521,7 @@ int adapter_start(struct btd_adapter *adapter)
 
 	memcpy(dev->features, features, 8);
 
-	if (hci_read_class_of_dev(dd, dev->class, 1000) < 0) {
+	if (hci_read_class_of_dev(dd, dev->class, HCI_REQ_TIMEOUT) < 0) {
 		err = errno;
 		error("Can't read class of adapter on %s: %s (%d)",
 					adapter->path, strerror(err), err);
@@ -2391,7 +2529,7 @@ int adapter_start(struct btd_adapter *adapter)
 		return -err;
 	}
 
-	if (hci_read_local_name(dd, sizeof(name), name, 2000) < 0) {
+	if (hci_read_local_name(dd, sizeof(name), name, HCI_REQ_TIMEOUT) < 0) {
 		err = errno;
 		error("Can't read local name on %s: %s (%d)",
 					adapter->path, strerror(err), err);
@@ -2405,9 +2543,10 @@ int adapter_start(struct btd_adapter *adapter)
 		goto setup;
 
 	if (ioctl(dd, HCIGETAUTHINFO, NULL) < 0 && errno != EINVAL)
-		hci_write_simple_pairing_mode(dd, 0x01, 2000);
+		hci_write_simple_pairing_mode(dd, 0x01, HCI_REQ_TIMEOUT);
 
-	if (hci_read_simple_pairing_mode(dd, &dev->ssp_mode, 1000) < 0) {
+	if (hci_read_simple_pairing_mode(dd, &dev->ssp_mode,
+						HCI_REQ_TIMEOUT) < 0) {
 		err = errno;
 		error("Can't read simple pairing mode on %s: %s (%d)",
 					adapter->path, strerror(err), err);
@@ -2453,7 +2592,8 @@ static void reply_pending_requests(struct btd_adapter *adapter)
 
 		if (adapter->bonding->io_id)
 			g_source_remove(adapter->bonding->io_id);
-		g_io_channel_close(adapter->bonding->io);
+		if (adapter->bonding->io)
+			g_io_channel_close(adapter->bonding->io);
 		bonding_request_free(adapter->bonding);
 		adapter->bonding = NULL;
 	}
@@ -2486,6 +2626,7 @@ static void unload_drivers(struct btd_adapter *adapter)
 int adapter_stop(struct btd_adapter *adapter)
 {
 	const char *mode = "off";
+	gboolean powered, discoverable;
 
 	/* cancel pending timeout */
 	if (adapter->discov_timeout_id) {
@@ -2531,6 +2672,20 @@ int adapter_stop(struct btd_adapter *adapter)
 					ADAPTER_INTERFACE, "Mode",
 					DBUS_TYPE_STRING, &mode);
 
+	powered = FALSE;
+
+	dbus_connection_emit_property_changed(connection, adapter->path,
+					ADAPTER_INTERFACE, "Powered",
+					DBUS_TYPE_BOOLEAN, &powered);
+
+	if (adapter->scan_mode == (SCAN_PAGE | SCAN_INQUIRY)) {
+		discoverable = FALSE;
+
+		dbus_connection_emit_property_changed(connection, adapter->path,
+					ADAPTER_INTERFACE, "Discoverable",
+					DBUS_TYPE_BOOLEAN, &discoverable);
+	}
+
 	adapter->up = 0;
 	adapter->scan_mode = SCAN_DISABLED;
 	adapter->mode = MODE_OFF;
@@ -2673,7 +2828,7 @@ const gchar *adapter_get_path(struct btd_adapter *adapter)
 
 void adapter_get_address(struct btd_adapter *adapter, bdaddr_t *bdaddr)
 {
-	str2ba(adapter->address, bdaddr);
+	bacpy(bdaddr, &adapter->bdaddr);
 }
 
 static gboolean discov_timeout_handler(void *data)
@@ -2703,7 +2858,7 @@ static gboolean discov_timeout_handler(void *data)
 	rq.rlen   = sizeof(status);
 	rq.event  = EVT_CMD_COMPLETE;
 
-	if (hci_send_req(dd, &rq, 1000) < 0) {
+	if (hci_send_req(dd, &rq, HCI_REQ_TIMEOUT) < 0) {
 		error("Sending write scan enable command to hci%d failed: %s (%d)",
 				dev_id, strerror(errno), errno);
 		goto failed;
@@ -2934,17 +3089,33 @@ void adapter_mode_changed(struct btd_adapter *adapter, uint8_t scan_mode)
 {
 	const char *mode;
 	const gchar *path = adapter_get_path(adapter);
-
-	adapter_set_scan_mode(adapter, scan_mode);
+	gboolean powered;
+	gboolean discoverable;
 
 	switch (scan_mode) {
 	case SCAN_DISABLED:
 		mode = "off";
 		adapter_set_mode(adapter, MODE_OFF);
+		powered = FALSE;
+		dbus_connection_emit_property_changed(connection, path,
+						ADAPTER_INTERFACE, "Powered",
+						DBUS_TYPE_BOOLEAN, &powered);
 		break;
 	case SCAN_PAGE:
 		mode = "connectable";
 		adapter_set_mode(adapter, MODE_CONNECTABLE);
+
+		if (adapter->scan_mode == SCAN_DISABLED) {
+			powered = TRUE;
+			dbus_connection_emit_property_changed(connection, path,
+						ADAPTER_INTERFACE, "Powered",
+						DBUS_TYPE_BOOLEAN, &powered);
+		} else {
+			discoverable = FALSE;
+			dbus_connection_emit_property_changed(connection, path,
+						ADAPTER_INTERFACE, "Discoverable",
+						DBUS_TYPE_BOOLEAN, &discoverable);
+		}
 		break;
 	case (SCAN_PAGE | SCAN_INQUIRY):
 
@@ -2957,6 +3128,17 @@ void adapter_mode_changed(struct btd_adapter *adapter, uint8_t scan_mode)
 			adapter_set_mode(adapter, MODE_DISCOVERABLE);
 			mode = "discoverable";
 		}
+
+		if (adapter->scan_mode == SCAN_DISABLED) {
+			powered = TRUE;
+			dbus_connection_emit_property_changed(connection, path,
+						ADAPTER_INTERFACE, "powered",
+						DBUS_TYPE_BOOLEAN, &powered);
+		}
+		discoverable = TRUE;
+		dbus_connection_emit_property_changed(connection, path,
+						ADAPTER_INTERFACE, "Discoverable",
+						DBUS_TYPE_BOOLEAN, &discoverable);
 		break;
 	case SCAN_INQUIRY:
 		/* Address the scenario where another app changed the scan mode */
@@ -2972,6 +3154,8 @@ void adapter_mode_changed(struct btd_adapter *adapter, uint8_t scan_mode)
 	dbus_connection_emit_property_changed(connection, path,
 					ADAPTER_INTERFACE, "Mode",
 					DBUS_TYPE_STRING, &mode);
+
+	adapter_set_scan_mode(adapter, scan_mode);
 }
 
 struct agent *adapter_get_agent(struct btd_adapter *adapter)
@@ -3047,7 +3231,8 @@ void adapter_free_bonding_request(struct btd_adapter *adapter)
 	if (adapter->bonding->io_id)
 		g_source_remove(adapter->bonding->io_id);
 
-	g_io_channel_close(adapter->bonding->io);
+	if (adapter->bonding->io)
+		g_io_channel_close(adapter->bonding->io);
 
 	bonding_request_free(adapter->bonding);
 
@@ -3056,7 +3241,7 @@ void adapter_free_bonding_request(struct btd_adapter *adapter)
 
 struct bonding_request_info *adapter_get_bonding_info(struct btd_adapter *adapter)
 {
-	if (!adapter || !adapter->bonding)
+	if (!adapter)
 		return NULL;
 
 	return adapter->bonding;
@@ -3176,3 +3361,22 @@ int btd_cancel_authorization(const bdaddr_t *src, const bdaddr_t *dst)
 
 	return agent_cancel(agent);
 }
+
+void adapter_update_devices(struct btd_adapter *adapter)
+{
+	char **devices;
+	int i;
+	GSList *l;
+
+	/* Devices */
+	devices = g_new0(char *, g_slist_length(adapter->devices) + 1);
+	for (i = 0, l = adapter->devices; l; l = l->next, i++) {
+		struct btd_device *dev = l->data;
+		devices[i] = (char *) device_get_path(dev);
+	}
+
+	dbus_connection_emit_property_changed(connection, adapter->path,
+					ADAPTER_INTERFACE, "Devices",
+					DBUS_TYPE_ARRAY, &devices);
+	g_free(devices);
+}
